package com.jh.common.jpa;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.function.Function;

import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;
import javax.persistence.criteria.CriteriaQuery;

import org.hibernate.Criteria;
import org.hibernate.criterion.Conjunction;
import org.hibernate.criterion.Disjunction;
import org.hibernate.criterion.MatchMode;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.ProjectionList;
import org.hibernate.criterion.Projections;
import org.hibernate.criterion.Restrictions;
import org.hibernate.query.NativeQuery;
import org.hibernate.transform.Transformers;

import com.jh.common.web.list.Page;
import com.jh.common.web.list.PagingInfo;
import com.jh.common.web.list.SortTrend;
import com.jh.common.web.list.SortingInfo;


/**
 * TODO comment.
 *
 * @author jh
 */
public abstract class AbstractHibernateRepository {

    /** The session factory. */
    @PersistenceContext
    protected EntityManager entityManager;



    /**
     * Criteria.
     *
     * @param clazz the clazz
     * @return the criteria
     */
    protected <T> CriteriaQuery<T> criteria(Class<T> clazz) {
    	return entityManager.getCriteriaBuilder().createQuery(clazz);
    }
    
    
    @SuppressWarnings("unchecked")
	protected <T> NativeQuery<T> nativeQuery(String query) {
    	return (NativeQuery<T>) entityManager.createNativeQuery(query);
    }
    
    @SuppressWarnings("unchecked")
	protected <T> T singleResult(Query q) {
    	try {
    		return (T) q.getSingleResult();
    	} catch (NoResultException e) {
    		return null;
    	}
    }
    
    /**
     * TODO comment
     *  
     *
     * @param tokens the tokens
     * @param fields the fields
     * @return the conjunction
     */
    protected Conjunction fulltextCriteria(List<String> tokens, String...fields) {
        Objects.requireNonNull(tokens, "Tokens must not be null");
        Objects.requireNonNull(fields, "Fields must not be null");
        Conjunction and = Restrictions.conjunction();
        for (String t : tokens) {
            Disjunction d = Restrictions.disjunction();
            for (String f : fields) {
                d.add(Restrictions.ilike(f, t, MatchMode.ANYWHERE));
            }
            and.add(d);
        }
        return and;
    }

    
    /**
     * TODO comment.
     *
     * @param c the c
     * @param clazz the clazz
     * @param properties the properties
     */
    protected void fields(Criteria c, Class<?> clazz, String...properties) {
        Objects.requireNonNull(c, "Criteria must not be null");
        Objects.requireNonNull(clazz, "Class must not be null");
        
        if (properties == null || properties.length == 0) {
            return;
        }
        
        ProjectionList pl = Projections.projectionList();
        for (String p : properties) {
            pl.add(Projections.alias(Projections.property(p), p));
        }
        c.setProjection(pl);
        c.setResultTransformer(Transformers.aliasToBean(clazz));
    }

    
    /**
     * TODO comment.
     *
     * @param <T> the generic type
     * @param c the c
     * @param pagingInfo the paging info
     * @return the page
     */
	@SuppressWarnings("unchecked")
	protected <T> Page<T> pagedResult(Criteria c, PagingInfo pagingInfo) {
		if (pagingInfo == null) {
			return new Page<>(c.list());
		} else if (pagingInfo.getPageSize() < 1) {
			return new Page<>(Collections.emptyList());
		}
		
		int pageSize = pagingInfo.getPageSize();
		
		int count = ((Number) c.setProjection(Projections.rowCount()).uniqueResult()).intValue();
		
		//count pages count
		int pagesCount = count / pageSize + ( count % pageSize == 0 ? 0 : 1);
		
		//limit results
		c.setFirstResult(pageSize * pagingInfo.getPage());
		c.setMaxResults(pageSize);
		
		//reset criteria https://stackoverflow.com/a/1472958
		c.setProjection(null);
		c.setResultTransformer(Criteria.ROOT_ENTITY);
		
		sortedCriteria(c, pagingInfo);
		
		return new Page<>(pagingInfo.getPage(), pagesCount, c.list(), pagingInfo.getField(), pagingInfo.getTrend());
	}

	
	/**
	 * Sorted criteria.
	 *
	 * @param c the c
	 * @param sortingInfo the sorting info
	 * @return the criteria
	 */
	protected Criteria sortedCriteria(Criteria c, SortingInfo sortingInfo) {
		if (sortingInfo != null && sortingInfo.getField() != null) {
			c.addOrder(sortingInfo.getTrend() == SortTrend.DESCENDING ? Order.desc(sortingInfo.getField()) : Order.asc(sortingInfo.getField()));
		}
		return c;
	}
	
	
	/**
	 * TODO.
	 *
	 * @param <T> the generic type
	 * @param entity the entity
	 */
    public <T extends AbstractIdEntity> void save(T entity) {
        if (entity.getId() == null) {
            entityManager.persist(entity);
        } else {
            entityManager.merge(entity);
        }
    }

    /**
     * Delete.
     *
     * @param <T> the generic type
     * @param entity the entity
     */
    public <T extends AbstractIdEntity> void delete(T entity) {
        if (entity.getId() != null) {
        	entityManager.remove(entity);
        }
    }

	/**
	 * Sync references.
	 *
	 * @param <T> the generic type
	 * @param <S> the generic type
	 * @param entity the entity
	 * @param referencedClass the referenced class
	 * @param referencedField the referenced field
	 * @param references the references
	 */
	//FIXME??
	protected <T extends AbstractIdEntity, S extends AbstractIdEntity> void syncReferences(T entity, String referencedClass, String referencedField, Collection<S> references) {
		if (entity.getId() == null) {
			//nothing to do
			return;
		}
		
		//filter only already stored references
		Set<S> elements = new HashSet<>();
		if (references != null) {
			references.stream().filter(e -> e.getId() != null).forEach(e -> elements.add(e));
		}
		
		StringBuilder sb = new StringBuilder("delete from ");
		sb.append(referencedClass).append(" ref");
		sb.append(" where ref.").append(referencedField).append(" = :root");
		
		if (!elements.isEmpty()) {
			sb.append(" and ref not in (:refs)");
		}
		
    	Query q = entityManager.createQuery(sb.toString());
    	q.setParameter("root", entity);
    	if (!elements.isEmpty()) {
    		q.setParameter("refs", elements);
    	}

    	q.executeUpdate();
    	if (references != null) {
        	for (S s : references) {
        	    save(s);
        	}
    	}
	}
	
	
	/**
	 * Sync references.
	 *
	 * @param <T> the generic type
	 * @param <S> the generic type
	 * @param entities the entities
	 * @param referencedClass the referenced class
	 * @param referencedField the referenced field
	 * @param valueProviders the value providers
	 */
	protected <T extends AbstractIdEntity, S extends AbstractIdEntity> void syncReferences(Collection<T> entities, String referencedClass, String referencedField, Function<T, ? extends Collection<S>> valueProviders) {
		if (entities == null || entities.isEmpty()) {
			//nothing to do
			return;
		}
		
		for (T entity : entities) {
			if (entity.getId() == null) {
				continue;
			}
			syncReferences(entity, referencedClass, referencedField, valueProviders.apply(entity));
		}
	}
	
	
//	@SuppressWarnings("unchecked")
//    public <T extends AbstractIdEntity> T merge(T entity) {
//		return (T) getSession().merge(entity);
//	}
	
}
