package com.jh.common.security;

import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.spec.InvalidKeySpecException;
import java.util.Base64;
import java.util.Base64.Decoder;
import java.util.Base64.Encoder;

import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.PBEKeySpec;

import org.springframework.util.StringUtils;

/**
 * The Class PasswordSecurity.
 */
public abstract class PasswordUtils {

	/** The Constant ITERATIONS. */
	private static final int ITERATIONS = 1;
	
	/** The b 64 encoder. */
	private static Encoder b64Encoder = Base64.getEncoder();
	
	/** The b 64 decoder. */
	private static Decoder b64Decoder = Base64.getDecoder();

	
	/**
	 * Hash password.
	 *
	 * @param password the password
	 * @return the password hash
	 */
	public static PasswordHash hashPassword(String password) {
	    try {
	    	byte[] salt = getSalt();
	    	String hash = createHash(password, salt);
			String salt64 = new String(b64Encoder.encode(salt));
			
			return new PasswordHash(hash, salt64);
			
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}


	/**
	 * Check password.
	 *
	 * @param password the password
	 * @param storedHash the stored hash
	 * @param storedSalt the stored salt
	 * @return true, if successful
	 */
	public static boolean checkPassword(String password, String storedHash, String storedSalt) {
		if (password == null || storedSalt == null || storedHash == null) {
			return false;
		}
		byte[] salt = b64Decoder.decode(storedSalt);
		try {
			String hash = createHash(password, salt);
			
			return hash.equals(storedHash);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	
	/**
	 * Validate password.
	 *
	 * @param password the password
	 * @return true, if successful
	 */
	public static boolean validatePassword(String password) {
		return StringUtils.hasText(password);
	}

	/**
	 * Creates the hash.
	 *
	 * @param password the password
	 * @param salt the salt
	 * @return the string
	 * @throws NoSuchAlgorithmException the no such algorithm exception
	 * @throws InvalidKeySpecException the invalid key spec exception
	 */
	protected static String createHash(String password, byte[] salt) throws NoSuchAlgorithmException, InvalidKeySpecException {
		PBEKeySpec spec = new PBEKeySpec(password.toCharArray(), salt, ITERATIONS, 2048);
		SecretKeyFactory factory = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA512");
		String hash = new String(b64Encoder.encode(factory.generateSecret(spec).getEncoded()));
		spec.clearPassword();
		return hash;
	}	
	
	/**
	 * Gets the salt.
	 *
	 * @return the salt
	 * @throws NoSuchAlgorithmException the no such algorithm exception
	 */
	private static byte[] getSalt() throws NoSuchAlgorithmException {
	    //Always use a SecureRandom generator
	    SecureRandom sr = SecureRandom.getInstance("SHA1PRNG");
	    //Create array for salt
	    byte[] salt = new byte[128];
	    //Get a random salt
	    sr.nextBytes(salt);
	    //return salt
	    return salt;
	}
	
}
