/*
 * Created on 27. 8. 2017 20:53:14
 *
 */
package com.jh.common.utils;

import java.time.LocalDate;
import java.time.LocalTime;
import java.time.Period;
import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.time.temporal.Temporal;
import java.time.temporal.TemporalAdjusters;
import java.time.temporal.TemporalUnit;
import java.util.Collections;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.TimeZone;

/**
 * The Class DateUtils.
 *
 * @author Jan Hrabal
 */
public class DateUtils {

	//private Map<>
	
	
	/**
	 * Start of.
	 *
	 * @param <T> the generic type
	 * @param t the t
	 * @param period the period
	 * @param unit the unit
	 * @return the t
	 */
	@SuppressWarnings("unchecked")
    public static <T extends Temporal> T startOf(T t, int period, ChronoUnit unit) {
		if (period > 1) {
			t = unit.addTo(t, -period + 1);
		}
		
		if (unit == ChronoUnit.MONTHS) {
			return (T) t.with(TemporalAdjusters.firstDayOfMonth());
			
		} else if (unit == ChronoUnit.YEARS) {
			return (T) t.with(TemporalAdjusters.firstDayOfYear());
		} 
		
		return t;
	}
	

	/**
	 * End of.
	 *
	 * @param <T> the generic type
	 * @param t the t
	 * @param period the period
	 * @param unit the unit
	 * @return the t
	 */
	@SuppressWarnings("unchecked")
    public static <T extends Temporal> T endOf(T t, int period, ChronoUnit unit) {
		if (period > 1) {
			t = startOf(t, period, unit);
			t = unit.addTo(t, period);
		}
		
		if (unit == ChronoUnit.MONTHS) {
			t = unit.addTo(t, -1);
			return (T) t.with(TemporalAdjusters.lastDayOfMonth());
			
		} else if (unit == ChronoUnit.YEARS) {
			return (T) t.with(TemporalAdjusters.lastDayOfYear());
		} 
		
		return t;
	}
	
	
	/**
	 * Start of quarter.
	 *
	 * @param year the year
	 * @param q the q
	 * @return the local date
	 */
	public static LocalDate startOfQuarter(int year, int q) {
		if (q < 1 || q > 4) {
			throw new IllegalArgumentException("Use quarter 1-4");
		}

		return LocalDate.of(year, q * 3 - 2, 1);
	}
	
	
	/**
	 * Months between.
	 *
	 * @param from the from
	 * @param to the to
	 * @return the long
	 */
	public static long monthsBetween(LocalDate from, LocalDate to) {
	    if (from == null || to == null) {
	        return 1;
	    }
	    Period p = Period.between(from, to.plusDays(1));
	    long months = Math.abs(p.toTotalMonths());
	    
	    if (Math.abs(p.getDays()) > 1) {
	        months++;
	    }
	    
	    return Math.max(months, 1);
	}
	
	
	/**
	 * Periods.
	 *
	 * @param from the from
	 * @param to the to
	 * @param unit the unit
	 * @return the list
	 */
	public static List<LocalDate> periods(LocalDate from, LocalDate to, TemporalUnit unit) {
	    if (from == null || to == null) {
	        return Collections.emptyList();
	    }
	    LinkedList<LocalDate> periods = new LinkedList<>();
	    LocalDate date = to;
	    
	    int i = 1;
	    
	    while (date.compareTo(from) > 0) {
	        periods.addFirst(date);
            date = to.minus(i, unit).with(TemporalAdjusters.lastDayOfMonth());
            i++;
	    }
	    
	    return periods;
	}
	
	
	public static Date utcDate(LocalDate date, LocalTime time, TimeZone tz) {
		if (tz == null) {
			tz = TimeZone.getTimeZone("GMT");
		}
		ZonedDateTime zdt = ZonedDateTime.of(date,  time, tz.toZoneId());
		return Date.from(zdt.toInstant());
	}

	public static Date utcDate(LocalDate date, LocalTime time, String tzId) {
		return utcDate(date, time, getTimeZone(tzId));
	}

	
	public static TimeZone getTimeZone(String id) {
		if (id == null) {
			return TimeZone.getTimeZone("GMT");
		}
		
		return TimeZone.getTimeZone(id);
	}
}
