/*
 * Created on 3. 11. 2016 9:19:23
 *
 */
package com.jh.common.web.list;

import java.util.List;
import java.util.Objects;

import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.util.StringUtils;
import org.springframework.web.context.request.NativeWebRequest;

import com.jh.common.utils.Utils;


/**
 * The Class ListHelper.
 */
public class ListHelper {

    /** The Constant DEFAULT_PAGE_SIZE. */
    public static final int DEFAULT_PAGE_SIZE = 20;
    
    /**
     * Read paging info.
     *
     * @param request the request
     * @param defaultField the default field
     * @param defaultTrend the default trend
     * @return the paging info
     */
    public static PagingInfo readPagingInfo(NativeWebRequest request, String defaultField, SortTrend defaultTrend) {
        if (request == null) {
            return new DefaultPagingInfo();
        }
        
        Integer page = 0;
        Integer pageSize = 10;
        
		if (Utils.isTrue(request.getHeader("X-Paging"))) {
	        String s = request.getHeader("X-Paging-Page");
	        page = Utils.parseInt(s, 0);
	        s = request.getHeader("X-Paging-PageSize");
	        pageSize = Utils.parseInt(s, 10);
		}
        
        SortingInfo si = readSortingInfo(request, defaultField, defaultTrend);
        
        return si == null ? new DefaultPagingInfo(page, pageSize) : new DefaultPagingInfo(page, pageSize, si.getField(), si.getTrend());
    }
    
    
    public static PagingInfo sortedPagingInfo(PagingInfo pi, String sortField, SortTrend trend) {
    	return new DefaultPagingInfo(pi.getPage(), pi.getPageSize(), sortField, trend);
    }
    
    /**
     * Read paging info.
     *
     * @param request the request
     * @return the paging info
     */
    public static PagingInfo readPagingInfo(NativeWebRequest request) {
    	return readPagingInfo(request, null, null);
    }
    
    
	/**
	 * Read sorting info.
	 *
	 * @param request the request
	 * @param defaultField the default field
	 * @param defaultTrend the default trend
	 * @return the sorting info
	 */
	public static SortingInfo readSortingInfo(NativeWebRequest request, String defaultField, SortTrend defaultTrend) {
		if (request == null) {
			return null;
		}
		
		if (!(Utils.isTrue(request.getHeader("X-Sorting")) || defaultField != null)) {
			return null;
		}

		String field = null;
		String trend = null;
		SortTrend st = null;

		field = request.getHeader("X-Sorting-Field");
		trend = request.getHeader("X-Sorting-Trend");

		if (!StringUtils.hasText(field) && StringUtils.hasText(defaultField)) {
			field = defaultField;
		}

		if (trend == null) {
			if (defaultTrend != null) {
				st = defaultTrend;
			}
		} else {
			st = "desc".equalsIgnoreCase(trend) ? SortTrend.DESCENDING : SortTrend.ASCENDING;
		}

		if (!StringUtils.hasText(field) || st == null) {
			return null;
		}

		return new DefaultSortingInfo(field, st);
	}
	
    
    /**
     * Write paging info.
     *
     * @param page the page
     * @return the http headers
     */
    public static HttpHeaders writePagingInfo(Page<?> page) {
        return writePagingInfo(new HttpHeaders(), page);
    }

    
    /**
     * Writes complete headers for provided page .
     *
     * @param headers the headers
     * @param page the page
     * @return the http headers
     */
    public static HttpHeaders writePagingInfo(HttpHeaders headers, Page<?> page) {
        if (page == null) {
            return headers;
        }

        headers.add("X-Paging", "true");
        headers.add("X-Paging-Page", String.valueOf(page.getPage()));
        headers.add("X-Paging-PagesCount", String.valueOf(page.getPagesCount()));
        
    	writeSortingInfo(headers, page);
        
        return headers;
    }

    
    /**
     * Writes only sorting information for provided page.
     *
     * @param headers the headers
     * @param page the page
     * @return the http headers
     */
    public static HttpHeaders writeSortingInfo(HttpHeaders headers, Page<?> page) {
        if (page == null || !page.isSorted()) {
            return headers;
        }

        headers.add("X-Sorting", Objects.toString(page.isSorted()));
        if (page.isSorted()) {
	        headers.add("X-Sorting-Field", page.getSortBy());
	        headers.add("X-Sorting-Trend", page.getTrend() == SortTrend.DESCENDING ? "desc" : "asc");
        }

        return headers;
    }
    
    
    /**
     * Creates paged response entity.
     *
     * @param <T> the generic type
     * @param page the page
     * @return the response entity
     */
    public static <T> ResponseEntity<List<T>> pagedResponse(Page<T> page) {
    	return new ResponseEntity<>(page.getContent(), ListHelper.writePagingInfo(page), HttpStatus.OK);
    }


}


/**
 * Default implementation for {@link PagingInfo}
 *
 * @author Jan Hrabal
 */
class DefaultPagingInfo implements PagingInfo {
    
    private int pageSize;
    
    private int page;

    private String field;
    
    private SortTrend sortTrend;
    
    public DefaultPagingInfo() {
        this(null, null);
    }
    
    public DefaultPagingInfo(Integer page, Integer pageSize) {
    	this(page, pageSize, null, null);
    }

    public DefaultPagingInfo(Integer page, Integer pageSize, String field, SortTrend sortTrend) {
        super();
        this.pageSize = pageSize == null || pageSize < 1 ? 20 : pageSize;
        this.page = page == null ? 0 : page;
        this.field = field;
        this.sortTrend = sortTrend;
    }
    
    
	@Override
    public int getPageSize() {
        return pageSize;
    }

    @Override
    public int getPage() {
        return page;
    }

	@Override
	public String getField() {
		return field;
	}

	@Override
	public SortTrend getTrend() {
		return sortTrend;
	}

}